% Simple Demo of Convolutional Network Learning and Inference
% Viren Jain, Massachusetts Institute of Technology <viren@mit.edu>
% Free for use and modification for any purpose other than use in a commercial application or product

function [W, B]=train_simplecn(input_image, label_image, num_of_hidden_layers, num_of_feature_maps, filter_size, training_iterations, minibatch_size)
	
	clf
		
	nonlinearity=inline('1./(1+exp(-x))'); % sigmoid function
	derivative_of_nonlinearity=inline('x.*(1-x)'); % derivative of sigmoid
	lossfunction=inline('(x-y).^2');
	derivative_of_lossfunction=inline('x-y'); % derivative of squared error
	
	% store the number of feature maps in each layer
	num_of_maps_in_layer{1}=1; % first layer encodes input, assume a single input map
	for i=1:num_of_hidden_layers
		num_of_maps_in_layer{i+1}=num_of_feature_maps; % hidden layers, each with the same number number of feature maps
	end
	num_of_maps_in_layer{num_of_hidden_layers+2}=1; % final layer encodes output, assume a single output map
	
	% initialize filter weights, biases, learning rates
	for l=2:num_of_hidden_layers+2
		for feature_map=1:num_of_maps_in_layer{l}
			W{l, feature_map}=0.1*single(randn(filter_size, filter_size, num_of_maps_in_layer{l-1})); % other hidden layers
			B{l, feature_map}=0.1*randn;
			etaW{l}=0.1; etaB{l}=0.1;
		end
	end
	
	if(num_of_hidden_layers>1), etaW{num_of_hidden_layers+2}=0.001;, etaB{num_of_hidden_layers+2}=0.001;, end
	
	% due to the effect of multiple valid convolutions, we have to compute the size of input patch that will generate the desired size of output patch
	minibatch_input_size=minibatch_size+(2*(floor(filter_size/2))*(num_of_hidden_layers+1));
	err=[];
	current_output=zeros(size(input_image),'single');
	tic
	for n=1:training_iterations
	
		% select a random input patch and its corresponding labels
		y=ceil(rand*(size(input_image,1)-minibatch_input_size(1)))+floor(minibatch_input_size(1)/2); x=ceil(rand*(size(input_image,2)-minibatch_input_size(2)))+floor(minibatch_input_size(2)/2);	
		training_image_coords=[y-floor(minibatch_input_size(1)/2)+1:y+floor(minibatch_input_size(1)/2); x-floor(minibatch_input_size(1)/2)+1:x+floor(minibatch_input_size(1)/2)];
		label_image_coords=[y-floor(minibatch_size(1)/2)+1:y+floor(minibatch_size(1)/2); x-floor(minibatch_size(1)/2)+1:x+floor(minibatch_size(1)/2)];
		training_input_patch=input_image(training_image_coords(1,:), training_image_coords(2,:));
		training_label_patch=label_image(label_image_coords(1,:), label_image_coords(2,:));
		
		% perform inference on current model
		activity=fwd_pass(W, B, num_of_maps_in_layer, nonlinearity, training_input_patch);	
		
		% compute gradient updates using cost function
		sensitivity=bkwd_pass(W, B, num_of_maps_in_layer, derivative_of_nonlinearity, derivative_of_lossfunction, activity, training_label_patch);
		[gradientW, gradientB]=gradient_pass(W, B, num_of_maps_in_layer, activity, sensitivity);
		
		% update model using gradient
		for l=2:num_of_hidden_layers+2
			for feature_map=1:num_of_maps_in_layer{l}
				W{l, feature_map}=W{l, feature_map}+(gradientW{l, feature_map}*etaW{l});
				B{l, feature_map}=B{l, feature_map}+(gradientB{l, feature_map}*etaB{l});
			end
		end
		
		% status and statistics
		current_output(label_image_coords(1,:),label_image_coords(2,:))=activity{size(W,1)};
		err(n)= sum(sum(lossfunction(training_label_patch, activity{size(W,1)})))/numel(training_label_patch);	
		
		if(rem(n,100)==0)
			toc
			subplot(1,3,1), plot(convn(err,ones([1 min(length(err), 100)]),'valid')/100);, title(['training error; iteration number = ', num2str(n)]);
			subplot(1,3,2), imagesc(input_image);, title('input image');
			if(rem(n,1000)==0)
				fprintf('computing forward pass on entire image...');, test_img_activity=fwd_pass(W,B, num_of_maps_in_layer, nonlinearity, input_image);, fprintf('done.\n')
				subplot(1,3,3), imagesc(test_img_activity{size(W,1)},[0 1]);, title(['output from iteration number = ', num2str(n)]);
			end
			drawnow;
			tic
		end
	end
end

function [activity]=fwd_pass(W, B, num_of_maps_in_layer, nonlinearity, input)
	activity{1}(:,:,1)=input;
	for l=2:size(W,1)
		for feature_map=1:num_of_maps_in_layer{l}
			activity{l}(:,:,feature_map)=nonlinearity(convn_fast(activity{l-1}, W{l, feature_map}, 'valid') + B{l,feature_map});
		end
	end
end

function [sensitivity]=bkwd_pass(W, B, num_of_maps_in_layer, derivative_of_nonlinearity, derivative_of_lossfunction, activity, training_label_patch)
%	sensitivity{size(W,1),1}=derivative_of_lossfunction(training_label_patch, activity{size(W,1)});
sensitivity{size(W,1),1}=derivative_of_lossfunction(training_label_patch, activity{size(W,1)}).*derivative_of_nonlinearity(activity{size(W,1)});
for l=size(W,1):-1:3
		sensitivity{l-1}=zeros(size(activity{l-1}),'single');
		for feature_map=1:num_of_maps_in_layer{l}
			sensitivity{l-1}=sensitivity{l-1}+convn_fast(sensitivity{l}(:,:,feature_map), flipdims(W{l, feature_map}));
		end
		sensitivity{l-1}=sensitivity{l-1}.*derivative_of_nonlinearity(activity{l-1});
	end
end

function [gradientW, gradientB]=gradient_pass(W, B, num_of_maps_in_layer, activity, sensitivity);	
	for l=2:size(W,1)
		for feature_map=1:num_of_maps_in_layer{l}
			gradientW{l, feature_map}=flipdims(xcorr4valid_freq(activity{l-1}, sensitivity{l}(:,:,feature_map)));
			gradientB{l, feature_map}=sum(sum(sensitivity{l}(:,:,feature_map)))/numel(sensitivity{l}(:,:,feature_map));
		end
	end
end

